from datetime import datetime
from typing import Any, Optional
from pydantic import BaseModel

class VerifiableAddress(BaseModel):
    id: str
    value: str
    verified: bool
    via: str
    status: str
    verified_at: datetime
    created_at: datetime
    updated_at: datetime

class Identity(BaseModel):
    id: str
    schema_id: str
    schema_url: str
    state: str
    state_changed_at: datetime
    traits: dict[str, str]
    verifiable_addresses: list[VerifiableAddress]
    metadata_public: Optional[dict[str, Any]]
    created_at: datetime
    updated_at: datetime
    organization_id: Optional[str]

class Device(BaseModel):
    id: str
    ip_address: str
    user_agent: str
    location: str

class AuthenticationMethod(BaseModel):
    method: str
    aal: str
    completed_at: datetime

class UserAuthResponse(BaseModel):
    id: str
    active: bool
    expires_at: datetime
    authenticated_at: datetime
    authenticator_assurance_level: str
    authentication_methods: list[AuthenticationMethod]
    issued_at: datetime
    identity: Identity
    devices: list[Device]


from typing import Callable, Optional, TypeVar, overload
import httpx


from test_infinity_stones.network_router.http_exception import HttpException
from test_infinity_stones.network_router.http_request import HttpRequest
from test_infinity_stones.network_router.http_response import HttpResponse

ResponseType = TypeVar('ResponseType')

class NetworkRouter:
    @overload
    async def execute_http_request(self, http_request: HttpRequest[ResponseType]) -> HttpResponse[ResponseType]:
        """Executes the provided HttpRequest and returns the response."""
        ...


    @overload
    async def execute_http_request(
        self, http_request: HttpRequest[ResponseType],
        batch_processor: Callable[[HttpRequest], list[HttpRequest]]
    ) -> HttpResponse[ResponseType]:
        """Executes the provided HttpRequest and returns the response."""
        ...


    async def execute_http_request(
        self,
        http_request: HttpRequest[ResponseType],
        batch_processor: Optional[Callable[[HttpRequest], list[HttpRequest]]] = None
    ) -> HttpResponse[ResponseType] | list[HttpResponse[ResponseType]]:
        """Executes the provided HttpRequest and returns the response."""

        if batch_processor:
            return await self._execute_batch_http_request(http_request, batch_processor)
        else:
            return await self._execute_plain_http_request(http_request)


    async def _execute_plain_http_request(
        self, http_request: HttpRequest[ResponseType]
    ) -> HttpResponse[ResponseType]:
        """Executes the provided HttpRequest and returns the response."""

        async with httpx.AsyncClient() as client:
            try:
                httpx_response = await client.request(
                    method=http_request.method.value,
                    url=http_request.url,
                    headers=http_request.headers,
                    params=http_request.query_params,
                    json=http_request.body,
                )
            except httpx.HTTPStatusError as e:
                raise HttpException(e.response.status_code, e.response.text)

        http_response = self.create_http_response(http_request, httpx_response)

        return http_response


    async def _execute_batch_http_request(
        self,
        http_request: HttpRequest,
        batch_processor: Callable[[HttpRequest], list[HttpRequest]]
    ) -> list[HttpResponse]:
        """Executes the provided HttpRequest and returns the response."""

        batched_http_requests = batch_processor(http_request)
        batched_http_responses: list[HttpResponse] = []

        for batched_http_request in batched_http_requests:
            batched_http_response = await self._execute_plain_http_request(batched_http_request)
            batched_http_responses.append(
                HttpResponse(
                    status_code=200,
                    body=batched_http_response,
                )
            )

        return batched_http_responses


    def create_http_response(
        self, http_request: HttpRequest[ResponseType], httpx_response: httpx.Response
    ) -> HttpResponse[ResponseType]:
        httpx_response_body = httpx_response.json()

        response_body = (
            http_request.response_body_decoder(httpx_response_body)
            if http_request.response_body_decoder and 200 <= httpx_response.status_code <= 299
            else httpx_response_body
        )

        http_response = HttpResponse(
            status_code=httpx_response.status_code,
            headers=httpx_response.headers.__dict__,
            body=response_body,
        )

        return http_response


import datetime
from enum import Enum
from typing import Any, Optional

from pydantic import BaseModel

class HttpMethod(Enum):
    """HTTP methods for HTTP requests"""

    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    PATCH = "PATCH"
    DELETE = "DELETE"
    HEAD = "HEAD"
    OPTIONS = "OPTIONS"
    TRACE = "TRACE"
    CONNECT = "CONNECT"

class APIAuth(BaseModel):
    """Authentication types for HTTP requests"""

    @staticmethod
    def basic(username: str, password: str) -> dict[str, str]:
        return {"Authorization": f"Basic {username}:{password}"}

    @staticmethod
    def api_key(api_key: str) -> dict[str, str]:
        return {"API-Key": api_key}

    @staticmethod
    def bearer(token: str) -> dict[str, str]:
        return {"Authorization": f"Bearer {token}"}

    @staticmethod
    def user_id_and_token(user_id: str, access_token: str) -> dict[str, str]:
        return {"user_id": f"{user_id}", "access_token": f"{access_token}"}

    @staticmethod
    def client_id_and_token(client_id: str, access_token: str) -> dict[str, str]:
        return {"client_id": f"{client_id}", "access_token": f"{access_token}"}

    @staticmethod
    def custom(custom_auth_headers: dict[str, Any]) -> dict[str, str]:
        return custom_auth_headers

    @staticmethod
    def none() -> dict[str, str]:
        return {}


from typing import Any, Generic, TypeVar
from pydantic import BaseModel

ResponseType = TypeVar("ResponseType")

class HttpResponse(BaseModel, Generic[ResponseType]):
    """
    Represents an HTTP response from an API endpoint.
    """

    status_code: int
    headers: Any = {}
    body: ResponseType


from test_infinity_stones.network_router.http_request import HttpRequest
from test_infinity_stones.network_router.http_response import HttpResponse
from test_infinity_stones.network_router.http_types import HttpMethod, APIAuth
from test_infinity_stones.network_router.router import NetworkRouter


from typing import Dict, Any, Generic, Optional, Callable, TypeVar
from pydantic import BaseModel
from test_infinity_stones.network_router.http_types import HttpMethod
from test_infinity_stones.network_router.utils.url_builder import build_url

ResponseType = TypeVar("ResponseType")

class HttpRequest(BaseModel, Generic[ResponseType]):
    """
    Represents an HTTP request to be made to an HTTP endpoint.
    """

    method: HttpMethod
    base_url: str
    path: str
    api_version: Optional[str] = None
    query_params: Optional[Dict[str, Any]] = None
    headers: Optional[Dict[str, str]] = None
    body: Optional[Dict[str, Any] | list[Any]] = None
    response_body_decoder: Optional[Callable[[Any], ResponseType]] = None
    request_body_encoder: Optional[Callable[[Any], Dict[str, Any]]] = None

    @property
    def url(self) -> str:
        """Constructs the full URL for the endpoint following on the `base_url/api_version/path` scheme"""
        return build_url(self.base_url, self.api_version, self.path)


from typing import Any

from fastapi import HTTPException, status
from test_infinity_stones.auth.user_auth_schemas import UserAuthResponse
from test_infinity_stones.network_router import HttpRequest, APIAuth, HttpMethod, NetworkRouter
from test_infinity_stones.config import Settings
import logging

logger = logging.getLogger()

class Authenticator:
    def __init__(self, settings: Settings, network_router: NetworkRouter = NetworkRouter()):
        self._settings = settings
        self._network_router = network_router


    async def authenticate_user(self, access_token: str) -> UserAuthResponse:
        request = HttpRequest[Any](
            method=HttpMethod.GET,
            base_url=self._settings.AUTH_SERVICE_BASE_URL,
            path=self._settings.AUTHENTICATE_USER_ENDPOINT_PATH,
            headers=APIAuth.bearer(access_token),
        )

        try:
            response = await self._network_router.execute_http_request(request)

            if response.status_code != 200:
                raise HTTPException(status_code=response.status_code, detail=response.body)

        except Exception as e:
            logger.error(f"Error while authenticating user {e}")
            raise HTTPException(status_code=status.HTTP_500_INTERNAL_SERVER_ERROR)

        return UserAuthResponse(**response.body)


import logging
from typing import Any, Callable, Optional, TypeVar
from uuid import uuid4
import redis.asyncio as redis

T = TypeVar("T")
U = TypeVar("U")

logger = logging.Logger(__name__)


class RedisClient:
    def __init__(self, client: redis.Redis):
        self.redis_client = client


    async def set_value(
        self,
        key: T,
        value: U,
        key_encoder: Optional[Callable[[Any], str]] = None,
        value_encoder: Optional[Callable[[Any], str]] = None,
        expiration_time: Optional[int] = None
    ):
        await self.set_values(
            keys_by_values={key: value},
            key_encoder=key_encoder,
            value_encoder=value_encoder,
            expiration_time=expiration_time
        )


    async def set_values(
        self: "RedisClient",
        keys_by_values: dict[T, U],
        key_encoder: Optional[Callable[[Any], str]] = None,
        value_encoder: Optional[Callable[[Any], str]] = None,
        expiration_time: Optional[int] = None
    ):
        pipeline = self.redis_client.pipeline()

        for key, value in keys_by_values.items():
            if value is None:
                continue

            string_key = self._encode_key_as_string(key, key_encoder)
            string_value = self._encode_value_as_string(value, value_encoder)

            pipeline.set(string_key, string_value, ex=expiration_time)

        await pipeline.execute()


    async def get_value(
        self: "RedisClient",
        key: str
    ) -> Optional[str]:
        result = await self.redis_client.get(key)
        return result


    async def _get_values(
        self: "RedisClient",
        keys: list[str]
    ) -> list[Optional[str]]:
        results = await self.redis_client.mget(keys)
        return results


    async def get_values_dict(
        self: "RedisClient",
        keys: list[str]
    ) -> dict[str, Optional[str]]:
        results = await self._get_values(keys)
        return self._create_dictionary(keys, results)


    async def get_decoded_value(
        self: "RedisClient",
        key: U,
        value_decoder: Callable[[str], T],
        key_encoder: Optional[Callable[[Any], str]] = None,
    ) -> Optional[T]:
        decoded_values = await self._get_decoded_values([key], value_decoder, key_encoder)

        if not decoded_values:
            return None

        return decoded_values[0]


    async def _get_decoded_values(
        self: "RedisClient",
        keys: list[T],
        value_decoder: Callable[[str], U],
        key_encoder: Optional[Callable[[Any], str]] = None,
    ) -> list[Optional[U]]:
        string_keys = [self._encode_key_as_string(key, key_encoder) or str(uuid4()) for key in keys]

        results = await self.redis_client.mget(string_keys)

        decoded_results: list[Optional[U]] = []

        for result in results:
            if result is None:
                decoded_results.append(None)
                continue

            decoded_result = self._decode_value(result, value_decoder)
            decoded_results.append(decoded_result)

        return decoded_results


    async def get_decoded_values_dict(
        self: "RedisClient",
        keys: list[T],
        value_decoder: Callable[[str], U],
        key_encoder: Optional[Callable[[T], str]] = None,
    ) -> dict[T, Optional[U]]:
        """Returns The values from cache as a dictionary whose keys are the original unencoded keys."""
        values = await self._get_decoded_values(keys, value_decoder, key_encoder)
        return self._create_dictionary(keys, values)


    async def delete(self: "RedisClient", key: T, key_encoder: Optional[Callable[[T], str]] = None):
        string_key = self._encode_key_as_string(key, key_encoder)

        self.redis_client.delete(string_key)


    async def clear_all_values(self):
        self.redis_client.flushdb()
        return True

    # MARK: Convenience Methods

    def _get_as_string(
        self, 
        value: Any,
        encoder: Optional[Callable[[Any], str]],
        no_encoder_error_message: str,
        invalid_encoder_return_type_error: str,
    ) -> str:
        if isinstance(value, str):
            return value

        if isinstance(value, int) or isinstance(value, float):
            return str(value)

        if encoder is None:
            raise ValueError(no_encoder_error_message)

        try:
            string_value = encoder(value)
        except Exception as e:
            logger.error(f"Error encoding value: {value}, error: {e}")
            raise e

        if not isinstance(string_value, str):
            raise ValueError(invalid_encoder_return_type_error)

        return string_value


    def _encode_key_as_string(self, key: Any, key_encoder: Optional[Callable[[Any], str]]) -> str:
        string_key = self._get_as_string(
            value=key,
            encoder=key_encoder,
            no_encoder_error_message=f"string key_encoder is required for non-string key: {key}",
            invalid_encoder_return_type_error=f"key encoder must return a string for key: {key}",
        )
    
        if string_key is None:
            raise ValueError(f"Error encoding key: {key}")

        return string_key


    def _encode_value_as_string(self, value: Any, value_encoder: Optional[Callable[[Any], str]]) -> str:
        return self._get_as_string(
            value=value,
            encoder=value_encoder,
            no_encoder_error_message=f"string value_encoder is required for non-string value: {value}",
            invalid_encoder_return_type_error=f"value encoder must return a string for value: {value}",
        )


    def _decode_value(self, value: str, value_decoder: Callable[[str], T]) -> T:
        try:
            decoded_result = value_decoder(value)
        except Exception as e:
            error_message = f"Redis client error decoding value: {value}, error: {e}"
            logger.error(error_message)
            raise Exception(error_message)
        return decoded_result


    def _create_dictionary(self, keys: list[Any], values: list[Any]) -> dict[Any, Any]:
        return dict(zip(keys, values))


async def get_redis_client(host: str, port: str, db: int) -> "RedisClient":
    url = f"redis://{host}:{port}/{db}"
    client = await redis.from_url(url, decode_responses=True)
    return RedisClient(client=client)